import { createLog } from '../log';
import type { Logger } from 'winston';
import { delay } from '../delay';

import { SerialPort, SerialPortMock } from 'serialport';
import { ReadlineParser } from '@serialport/parser-readline';
import { MockBinding } from '@serialport/binding-mock';

interface PortInfo {
    path: string;
    manufacturer: string | undefined;
    serialNumber: string | undefined;
    pnpId: string | undefined;
    locationId: string | undefined;
    productId: string | undefined;
    vendorId: string | undefined;
}

enum Commands {
	CONNECT = 'i',
	MCOPY_IDENTIFIER  = 'm',
	CAMERA_IDENTIFIER = 'k', 
	CAMERA = 'c',
	CAMERA_FORWARD = 'e',
	CAMERA_BACKWARD = 'f',
	CAMERA_OPEN = 'J',
	CAMERA_CLOSE = 'K'
}

class CameraSerialPortMock extends SerialPortMock {
	private log : Logger;
	private parser : ReadlineParser;
	constructor (options : any, openCallback : ErrorCallback = null, parser : ReadlineParser) {
		super(options, openCallback);
		this.log = createLog('mock');
		this.parser = parser;
	}

	write (buffer : any) : boolean {
		super.write(buffer)
		//this.log.info(this.prefix + `Received data: "${buffer}"`);
		switch (buffer) {
			case Commands.CONNECT :
				this._mockSend(Commands.CONNECT, 3);
				break;
			case Commands.MCOPY_IDENTIFIER :
				this._mockSend(Commands.CAMERA_IDENTIFIER, 2);
				break;
			case Commands.CAMERA :
				this._mockSend(Commands.CAMERA, 250);
				break;
			case Commands.CAMERA_OPEN :
				this._mockSend(Commands.CAMERA_OPEN, 125);
				break;
			case Commands.CAMERA_CLOSE :
				this._mockSend(Commands.CAMERA_CLOSE, 125);
				break;
			default:
				this.log.warn(`[MOCK] Command "${buffer}" does not exist on mock`);
		}
		return true;
	}

	_mockSend (buffer : string, delay : number) {
		setTimeout(function () {
			this.parser._events.data(buffer);
		}.bind(this), delay);
	}
}

export class Camera {
	private log : Logger;
	private parser : ReadlineParser; 
	private ready : boolean = false;
	private connected : boolean = false;
	private serial : SerialPort | CameraSerialPortMock = null;
	private baud : number = 57600;
	private next : Function = null;
	private port : string = null;
	private prefix : string = '';
	private mock : boolean = false;

	constructor (mock : boolean) {
		this.mock = mock;
		this.log = createLog('camera');
		this.parser = new ReadlineParser({ delimiter: '\r\n' });
		this.begin();
	}

	private async begin () {
		let ports : string[] = [];
		let selected : boolean = false;
		try {
			ports = await this.enumerate();
		} catch (err) {
			this.log.error(this.prefix + 'Error calling enumerate()', err);
		}
		if (!this.mock && ports.length > 0) {
			for (let port of ports) {
				this.log.info(this.prefix + `Found USB serial device: ${port} ${selected ? '*' : ''}`);
				selected = false;
			}
			try {
				await this.connect(ports[0]);
			} catch (err) {
				this.log.error(this.prefix + `Error connecting to ${ports[0]}`, err);
			}
		} else {
			if (this.mock) {
				this.log.info(`Starting camera in MOCK mode due to system setting`);
			} else {
				this.log.warn(this.prefix + `No USB serial devices found, connecting to MOCK...`)
			}
			try {
				await this.connectMock();
			} catch (err) {
				this.log.error(this.prefix + `Error connecting to MOCK USB serial device "/dev/fake"`, err);
			}
		}
		await delay(3000);
		await this.verify();
	}

	private filter (port : PortInfo) {
		if ((port.manufacturer + '').toLowerCase().indexOf('arduino') !== -1 ||
			(port.path + '').toLowerCase().indexOf('usbserial') !== -1 ||
			(port.path + '').toLowerCase().indexOf('usbmodem') !== -1 || 
			(port.path + '').toLowerCase().indexOf('ttyusb') !== -1) {
			return true;
		}
		return false;
	}

	private async enumerate () : Promise<string[]> {
		let listed : PortInfo[] = [];
		try {
			listed = await SerialPort.list();
		} catch (err) {
			this.log.error(this.prefix + 'Error listing serial ports', err);
		}
		return listed.filter(this.filter).map((port : PortInfo) => port.path);
	}

	private async connect (port : string) {
		this.port = port;
		try {
			this.serial = new SerialPort({
				path : this.port,
				baudRate: this.baud,
			});
		} catch (err) {
			this.log.error(this.prefix + 'Error creating SerialPort object', err);
			return;
		}
		if (this.serial !== null) {
			this.serial.pipe(this.parser);
			this.parser.on('data', this.onData.bind(this));
		}
		
		return new Promise(function (resolve : Function, reject : Function) {
			this.serial.on('open', () => {
  				this.log.info(this.prefix + `Connected to USB serial device ${this.port} @ ${this.baud} baud`);
  				this.connected = true;
  				return resolve(true);
			});
		}.bind(this));
	}

	private async connectMock () {
		this.port = '/dev/fake';
		try {
			SerialPortMock.binding.createPort(this.port);
			this.serial = new CameraSerialPortMock({ 
				path : this.port, 
				baudRate: this.baud 
			}, null, this.parser);
		} catch (err) {
			this.log.error(this.prefix + 'Error creating SerialPortMock object', err);
			return;
		}
		if (this.serial !== null) {
			this.serial.pipe(this.parser);
			this.parser.on('data', this.onData.bind(this));
		}
		return new Promise(function (resolve : Function, reject : Function) {
			this.serial.on('open', () => {
				this.prefix = '[MOCK] ';
  				this.log.info(this.prefix + `Connected to MOCK USB serial device ${this.port} @ ${this.baud} baud`);
  				this.connected = true;
  				return resolve(true);
			});
		}.bind(this));	
	}

	private onData (data : string) {
		this.log.info(this.prefix + `Received data: "${data}"`);
		if (this.next !== null) {
			this.next(data);
		}
	}

	private async verify () {
		try {
			await this.confirm(Commands.CONNECT, Commands.CONNECT);
			this.log.info(this.prefix + `Confirmed mcopy device`);
		} catch (err) {
			this.log.error(this.prefix + `Error connecting to mcopy device`, err);
			return;
		}
		try {
			await this.confirm(Commands.MCOPY_IDENTIFIER, Commands.CAMERA_IDENTIFIER);
			this.log.info(this.prefix + `Confirmed mcopy camera`);
		} catch (err) {
			this.log.error(this.prefix + `Error identifying device`, err);
			return;
		}
		this.ready = true;
		this.log.info(this.prefix + `Camera connected and ready`);
	}

	private async confirm (cmd : string, res : string) : Promise<boolean> {
		return new Promise (function (resolve : Function, reject: Function) {
			this.next = function (data : string) {
				this.next = null;
				if (data === res) {
					return resolve(true);
				} else {
					return reject(new Error(`Response ${data} !== ${res}`));
				}
			}.bind(this);
			this.log.info(this.prefix + `Send data: "${cmd}"`);
			this.serial.write(cmd, (err : any, results : any) => {
				if (err) {
					this.log.error(this.prefix + 'Error writing to device', err);
					return reject(err);
				}
			});
		}.bind(this));
	}

	public async frame () : Promise<number> {
		const start : number = Date.now();
		let ms : number;
		await this.confirm(Commands.CAMERA, Commands.CAMERA);
		ms = (Date.now()) - start;
		this.log.info(this.prefix + `frame() - ${ms}ms`);
		return ms;
	}

	public async open () : Promise<number> {
		const start : number = Date.now();
		let ms : number;
		await this.confirm(Commands.CAMERA_OPEN, Commands.CAMERA_OPEN);
		ms = (Date.now()) - start;
		this.log.info(this.prefix + `open() - ${ms}ms`);
		return ms;
	}

	public async close () : Promise<number> {
		const start : number = Date.now();
		let ms : number;
		await this.confirm(Commands.CAMERA_CLOSE, Commands.CAMERA_CLOSE);
		ms = (Date.now()) - start;
		this.log.info(this.prefix + `close() - ${ms}ms`);
		return ms;
	}

	public async direction (dir : boolean) : Promise<number> {
		const start : number = Date.now();
		let ms : number;
		const cmd : string = dir ? Commands.CAMERA_FORWARD : Commands.CAMERA_BACKWARD;
		await this.confirm(cmd, cmd);
		ms = (Date.now()) - start;
		this.log.info(this.prefix + `direction(${dir}) - ${ms}ms`);
		return ms;
	}
}

module.exports = { Camera };